from io import FileIO
from PIL import Image, ImageDraw

def ega_to_rgb(egaColor):
    red = int(255 * (0.6666 * (egaColor & 4) / 4 + 0.33333 * (egaColor & 8) / 8))
    green = int(255 * (0.6666 * (egaColor & 2) / 2 + 0.33333 * (egaColor & 8) / 8))
    blue = int(255 * (0.6666 * (egaColor & 1) / 1 + 0.33333 * (egaColor & 8) / 8))
    
    if egaColor == 6:
        green = int(green / 2)
        
    return (red, green, blue)
    
def load_tileset(fileName):
    file = FileIO(fileName)
    tileset = []
    
    while True:
        tiledata = file.read(128)
        if len(tiledata) <= 0:
            break
            
        tileImage = Image.new('RGB', (16, 16), 0)
        
        imageX, imageY = 0, 0
        for byte in tiledata:
            color1 = byte >> 4
            color2 = byte & 0xF
            
            tileImage.putpixel((imageX, imageY), ega_to_rgb(color1))
            tileImage.putpixel((imageX + 1, imageY), ega_to_rgb(color2))
            
            imageX += 2
            if imageX >= tileImage.width:
                imageX = 0
                imageY += 1
            
        tileset.append(tileImage)
    
    file.close()
    
    return tileset
    
def load_map(mapBytes, size, tileset):
    mapImage = Image.new('RGB', (16 * size[0], 16 * size[1]), 0)
    
    xCoord = 0
    yCoord = 0
    for tileId in mapBytes:
        mapImage.paste(tileset[tileId], (xCoord, yCoord))
        
        xCoord += 16
        if xCoord >= mapImage.width:
            xCoord = 0
            yCoord += 16

    return mapImage
    
def load_town(filename, talkFilename, tileset):
    town = {}

    townBytes = FileIO(filename).read()
    talkBytes = FileIO(talkFilename).read()
    
    town['map'] = load_map(townBytes[0x0:0x400], (32, 32), tileset)
    
    npcs = []
    for npcIndex in range(32):
        npc = {
            'tile': townBytes[0x400 + npcIndex],
            'x': townBytes[0x420 + npcIndex],
            'y': townBytes[0x440 + npcIndex],
            'conversationIndex': townBytes[0x4e0 + npcIndex]
        }
        
        if npc['x'] != 0 and npc['y'] != 0 and npc['tile'] != 0:
            if npc['conversationIndex'] != 0:
                npcNameChars = []
                npcNameStartByte = (npc['conversationIndex'] - 1) * 0x120 + 3
                npcNameEndByte = npcNameStartByte
                while npcNameEndByte < len(talkBytes) and talkBytes[npcNameEndByte] != 0:
                    npcNameEndByte += 1
                npc['name'] = talkBytes[npcNameStartByte:npcNameEndByte].decode('latin-1');
            npcs.append(npc)
    
    town['npcs'] = npcs
    
    return town
    
def draw_town(town, tileset):
    townImage = Image.new('RGB', town['map'].size, 0)
    townDrawInterface = ImageDraw.Draw(townImage)
    
    townImage.paste(town['map'])
    
    for npc in town['npcs']:
        townImage.paste(tileset[npc['tile']], (npc['x'] * 16, npc['y'] * 16))
        
    for npc in town['npcs']:
        if 'name' in npc:
            townDrawInterface.text((npc['x'] * 16, (npc['y'] + 1) * 16), npc['name'], 0x00ffff)
    
    return townImage
    
def load_dungeon(filename, tileset):
    dungeon = {}
    dungeonFile = FileIO(filename)
    
    maps = []
    for i in range(8):
        maps.append(dungeonFile.read(64))
    dungeon['maps'] = maps
    
    rooms = []
    while True:
        roomBytes = dungeonFile.read(256)
        if len(roomBytes) <= 0:
            break
            
        room = {}
        room['image'] = load_map(roomBytes[0x80:0xf9], (11, 11), tileset)
        
        for monsterIndex in range(16):
            if roomBytes[0x10 + monsterIndex] == 172: # Mimic
                room['image'].paste(tileset[175], (roomBytes[0x20 + monsterIndex] * 16, roomBytes[0x30 + monsterIndex] * 16))
            elif roomBytes[0x10 + monsterIndex] != 0:
                room['image'].paste(tileset[roomBytes[0x10 + monsterIndex]], (roomBytes[0x20 + monsterIndex] * 16, roomBytes[0x30 + monsterIndex] * 16))
                
        roomDrawInterface = ImageDraw.Draw(room['image'])
        for triggerIndex in range(4):
            targetTile = roomBytes[triggerIndex * 4]
            triggerX, triggerY = roomBytes[triggerIndex * 4 + 1] >> 4, roomBytes[triggerIndex * 4 + 1] & 0xF 
            target1X, target1Y = roomBytes[triggerIndex * 4 + 2] >> 4, roomBytes[triggerIndex * 4 + 2] & 0xF
            target2X, target2Y = roomBytes[triggerIndex * 4 + 3] >> 4, roomBytes[triggerIndex * 4 + 3] & 0xF
                           
            if targetTile != 0:
                roomDrawInterface.rectangle([triggerX * 16 + 1, triggerY * 16 + 1, (triggerX + 1) * 16 - 1, (triggerY + 1) * 16 - 1], outline=0x00ffff)
                
                if target1X != 0 or target1Y != 0:
                    roomDrawInterface.rectangle([target1X * 16, target1Y * 16, (target1X + 1) * 16 - 1, (target1Y + 1) * 16 - 1], outline=0x00ff00)
                    roomDrawInterface.line([(triggerX * 16 + 8, triggerY * 16 + 8), (target1X * 16 + 8, target1Y * 16 + 8)], 0x00ffff, 1)
                if target2X != 0 or target2Y != 0:
                    roomDrawInterface.rectangle([target2X * 16, target2Y * 16, (target2X + 1) * 16 - 1, (target2Y + 1) * 16 - 1], outline=0x00ff00)
                    roomDrawInterface.line([(triggerX * 16 + 8, triggerY * 16 + 8), (target2X * 16 + 8, target2Y * 16 + 8)], 0x00ffff, 1)
        
        rooms.append(room)
    dungeon['rooms'] = rooms
        
    return dungeon
    
def get_dungeon_cell(floorDataBytes, column, row):
    while column >= 8:
        column -= 8
    while column < 0:
        column += 8
    while row >= 8:
        row -= 8
    while row < 0:
        row += 8
        
    return floorDataBytes[row * 8 + column]
    
def draw_dungeon_floors(filename, tileset):
    dungeon = load_dungeon(filename, tileset)
    
    floorImages = []

    temp_double_ladder = tileset[27].copy()
    temp_double_ladder.paste(
        tileset[28].crop((0, temp_double_ladder.height // 4, temp_double_ladder.width, temp_double_ladder.height)), 
        (0, temp_double_ladder.height // 4)
    )
    dungeon_feature_tiles = {
        0x10: tileset[27],  #ladder up
        0x20: tileset[28],  #ladder down
        0x30: temp_double_ladder,   # ladder up/down
        0x40: tileset[60],  #treasure
        0x50: tileset[9], #ceiling hole
        0x60: tileset[9], #floor hole
        0x70: tileset[78], #magic orb
        0x80: tileset[115], #winds
        0x81: tileset[115], #rocks
        0x8e: tileset[115], #pit trap
        0x90: tileset[101], #water fountain
        0x91: tileset[101], #healing fountain
        0x92: tileset[101], #acid fountain
        0x93: tileset[101], #cure fountain
        0x94: tileset[101], #poison fountain
        0xa0: tileset[68], #poison field
        0xa1: tileset[69], #energy field
        0xa2: tileset[70], #fire field
        0xa3: tileset[71], #sleep field
        0xb0: tileset[74], #altar
        0xc0: tileset[59], #door
        0xe0: tileset[73], #secret door
    }
    
    dungeon_feature_annotations = {
        0x80: ('Wind', 0x00ffff),
        0x81: ('Rock', 0x0000ff),
        0x8e: ('Pit', 0x0000ff),
        0x90: ('Plain', 0x999999),
        0x91: ('Healing', 0xff0000),
        0x92: ('Acid', 0x0000ff),
        0x93: ('Cure', 0xff0000),
        0x94: ('Poison', 0x0000ff)
    }
    
    for floorIndex, map in enumerate(dungeon['maps']):
        xCoord = 0
        yCoord = 0
        floorImage = Image.new("RGB", (8 * 11 * 16, 8 * 11 * 16), 0)
        drawInterface = ImageDraw.Draw(floorImage)
        
        # Draw grid lines.
        for column in range(8):
            drawInterface.line((column * 11 * 16, 0, column * 11 * 16, 8 * 11 * 16), 0xff222222, 3)
            
        for row in range(8):
            drawInterface.line((0, row * 11 * 16, 8 * 11 * 16, row * 11 * 16), 0xff222222, 3)
        
        for index, byte in enumerate(map):
            left, top = xCoord, yCoord
            left_center, top_center = xCoord + 64, yCoord + 64
            right_center, bottom_center = xCoord + 112, yCoord + 112
            right, bottom = xCoord + 11 * 16, yCoord + 11 * 16
            
            row = int(index / 8)
            column = index % 8
            
            if byte >= 0xd0 and byte <= 0xdf:
                # Draw dungeon room.
                if len(dungeon['rooms']) > 16:
                    roomIndex = (byte - 0xd0) + ((floorIndex // 2) * 16)
                else:
                    roomIndex = byte - 0xd0
                room = dungeon['rooms'][roomIndex]
                floorImage.paste(room['image'], (xCoord, yCoord))
            elif byte != 0xf0:
            
                # Draw dungeon feature, if applicable.
                if byte in dungeon_feature_tiles:
                    floorImage.paste(dungeon_feature_tiles[byte].resize((right_center - left_center, bottom_center - top_center)), (left_center, top_center))
                
                # Annotate fountains and traps.
                if byte in dungeon_feature_annotations:
                    drawInterface.text((left_center + 2, bottom_center + 2), dungeon_feature_annotations[byte][0], dungeon_feature_annotations[byte][1])
            
                # Draw walls.
                if (get_dungeon_cell(map, column - 1, row) == 0xf0):
                    drawInterface.line([(left_center, top_center), (left_center, bottom_center)], 0xffffffff, 3)
                else:
                    drawInterface.line([(left, top_center), (left_center, top_center)], 0xffffffff, 3)
                    drawInterface.line([(left, bottom_center), (left_center, bottom_center)], 0xffffffff, 3)
                    
                if (get_dungeon_cell(map, column + 1, row) == 0xf0):
                    drawInterface.line([(right_center, top_center), (right_center, bottom_center)], 0xffffffff, 3)
                else:
                    drawInterface.line([(right, top_center), (right_center, top_center)], 0xffffffff, 3)
                    drawInterface.line([(right, bottom_center), (right_center, bottom_center)], 0xffffffff, 3)
                    
                if (get_dungeon_cell(map, column, row - 1) == 0xf0):
                    drawInterface.line([(left_center, top_center), (right_center, top_center)], 0xffffffff, 3)
                else:
                    drawInterface.line([(left_center, top), (left_center, top_center)], 0xffffffff, 3)
                    drawInterface.line([(right_center, top), (right_center, top_center)], 0xffffffff, 3)
                
                if (get_dungeon_cell(map, column, row + 1) == 0xf0):
                    drawInterface.line([(left_center, bottom_center), (right_center, bottom_center)], 0xffffffff, 3)
                else:
                    drawInterface.line([(left_center, bottom), (left_center, bottom_center)], 0xffffffff, 3)
                    drawInterface.line([(right_center, bottom), (right_center, bottom_center)], 0xffffffff, 3)
                
            
            xCoord += 11 * 16
            if xCoord >= floorImage.width:
                xCoord = 0
                yCoord += 11 * 16
                
        floorImages.append(floorImage)
        
    return floorImages

